/*
** Module   : ProxyAssembler.java
** Abstract : Assembles a dynamic proxy subclass.
**
** Copyright (c) 2010-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description------------------------
** 001 ECF 20101119 Created initial version.
** 002 ECF 20101216 Pass null to InvocationHandler.invoke() for arguments
**                  parameter, instead of an empty Object array, to match
**                  contract for this method. Generate different code, based
**                  on whether the superclass implements InvocationHandler.
** 003 ECF 20130122 Fix to enable nested interfaces to be used to define the
**                  proxy.
*/

package com.goldencode.proxy;

import java.lang.reflect.*;
import org.objectweb.asm.*;

/**
 * A class which assembles dynamic proxy classes using the ASM bytecode
 * engineering library (http://asm.ow2.org).  As a package private class, it
 * is intended to be invoked only by the {@link ProxyFactory}, which exposes
 * a public API for the definition of new proxy objects.
 * <p>
 * This mechanism is used in preference to the dynamic proxy implementation
 * exposed by <code>java.lang.reflect.Proxy</code> because the latter does not
 * allow one to specify a base class to extend.  In situations where a common
 * interface and implementation combination is used for many dynamically
 * specified classes, this can cause memory problems.  Specifically, the
 * amount of PermGen memory required for <code>java.lang.reflect.Method</code>
 * objects is duplicated across each proxy class which uses the common
 * interface/implementation, because the methods in the common interface are
 * duplicated in every proxy class.
 * <p>
 * The approach used by this assembler is to allow the specification of a base
 * class which implements the common interface, so that these common methods
 * are defined only in a single class, rather than duplicated across every
 * dynamically defined proxy class.  A user would define all common method
 * implementations in this base class, which each proxy class would then
 * inherit "for free".  All unique methods are still specified by an array of
 * interfaces passed to this factory's static <code>getProxy</code> method.
 * <p>
 * The architecture of this dynamic proxy implementation retains the use of
 * <code>java.lang.reflect.InvocationHandler</code>.  All calls to the methods
 * specified in the array of interfaces passed to <code>getProxy</code> are
 * routed to the invocation handler, as in the standard, dynamic proxy
 * implementation.  This requires that an <code>InvocationHandler</code>
 * instance is accessible to the runtime instance of the dynamically generated
 * class.  In the standard implementation, the base class of all proxy classes
 * is <code>java.lang.reflect.Proxy</code>, which holds a reference to the
 * invocation handler.  However, since there are no restrictions or contracts
 * regarding the base class in this implementation, the invocation handler
 * is stored in each proxy instance.  This adds a single, redundant field to
 * each proxy class definition, which is the trade-off for being able to
 * define all common methods in the base class of the user's choice.  So long
 * as there is at least one such common method in the base class to offset the
 * presence of this redundant field, this implementation should be more more
 * memory efficient than the standard proxy implementation.
 * <p>
 * Note:  the CGLib bytecode engineering library was considered as a simpler
 * alternative to this implementation;  however, the closest substitute in
 * that library (the Enhancer class) automatically overrides all non-final
 * methods of the base class in order to instrument callbacks in the
 * overridden methods.  This behavior was considered too heavyweight.  A
 * workaround would have been to require all such methods be marked as final
 * to avoid this override behavior;  however, this requirement was rejected.
 * <p>
 * The following pseudo-code representation is intended to convey the general
 * structure of the generated proxy class.  However, please note that no
 * intermediate source code representation is produced.
 * <p>
 * <pre>
 * package com.goldencode.proxy;
 * 
 * public final class [ProxyClassName]
 * implements [Interface List]
 * {
 *    // list of delegate methods
 *    private static Method[] ma;
 *    
 *    // invocation handler
 *    private final InvocationHandler h;
 *    
 *    public [ProxyClassName](InvocationHandler handler)
 *    {
 *       this.h = handler;
 *    }
 *    
 *    public static void __initializeMethods(Method[] methods)
 *    {
 *       if (ma != null)
 *       {
 *          throw new IllegalStateException(
 *             "Proxy methods may not be set more than once");
 *       }
 *       
 *       [ProxyClassName].ma = methods;
 *    }
 *    
 *    ...
 *    
 *    [implemented methods -- see inner class ProxyMethod for description]
 *    
 *    ...
 * }
 * </pre>
 * 
 * @author ECF
 */
final class ProxyAssembler
implements Constants
{
   /** Name of static method to initialize methods upon loading proxy class */
   static final String INIT_METHODS_METHOD = "__initializeMethods";
   
   /** Name of the proxy class to be assembled */
   private final String proxyName;
   
   /** Name of the proxy class' parent class */
   private final String parentName;
   
   /** Interfaces which the proxy class will implement */
   private final Class<?>[] interfaces;
   
   /** Method objects which will be passed to invocation handler */
   private final Method[] methods;
   
   /** Does parent class implement <code>InvocationHandler</code>? */
   private final boolean parentIsHandler;
   
   /** Object used to create class fields, methods, etc. */
   private final ClassWriter classWriter = new ClassWriter(false);
   
   /**
    * Constructor.
    * 
    * @param   proxyName
    *          Name of the proxy class to be assembled.  This is the external
    *          name known to the class loader.
    * @param   parentName
    *          Name of the proxy class' parent class.  This is the external
    *          name known to the class loader.
    * @param   interfaces
    *          Interfaces which the proxy class will implement.
    * @param   methods
    *          Methods to be proxied.
    * @param   parentIsHandler
    *          <code>true</code> if the parent class implements the
    *          <code>InvocationHandler</code> interface, else
    *          <code>false</code>.  If <code>true</code>, no separate
    *          instance field will be assembled to store the invocation
    *          handler reference, as the instance of this class itself will
    *          act as the handler.
    */
   ProxyAssembler(String proxyName,
                  String parentName,
                  Class<?>[] interfaces,
                  Method[] methods,
                  boolean parentIsHandler)
   {
      this.proxyName = Utils.commonToInternalTypeName(proxyName);
      this.parentName = Utils.commonToInternalTypeName(parentName);
      this.interfaces = interfaces;
      this.methods = methods;
      this.parentIsHandler = parentIsHandler;
   }
   
   /**
    * Assemble the class in accordance with the Java class file format as
    * specified in <u>The Java Virtual Machine Specification, Second
    * Edition</u> by Tim Lindholm and Frank Yellin.
    * <p>
    * See the class description above for details regarding the content of
    * the assembled class.
    * 
    * @return  Byte array containing the class definition.
    */
   byte[] assembleClass()
   {
      // gather the list of interfaces being implemented
      String[] ifaceNames = new String[interfaces.length];
      for (int i = 0; i < interfaces.length; i++)
      {
         Class<?> iface = interfaces[i];
         String name = iface.getName();
         ifaceNames[i] = Utils.commonToInternalTypeName(name);
      }
      
      // create the proxy class
      classWriter.visit(V1_5,
                        ACC_PUBLIC | ACC_FINAL,
                        proxyName,
                        parentName,
                        ifaceNames,
                        null);
      
      // add a class member field for the array of delegate Methods
      classWriter.visitField(ACC_PRIVATE | ACC_STATIC | ACC_FINAL,
                             "ma",
                             "[Ljava/lang/reflect/Method;",
                             null,
                             null);
      
      if (!parentIsHandler)
      {
         // add an instance member field for the InvocationHandler
         classWriter.visitField(ACC_PRIVATE | ACC_FINAL,
                                "h",
                                "Ljava/lang/reflect/InvocationHandler;",
                                null,
                                null);
      }
      
      // create the __initializeMethods method
      assembleInitializeMethods();
      
      // create a constructor
      if (parentIsHandler)
      {
         assembleDefaultConstructor();
      }
      else
      {
         assembleHandlerConstructor();
      }
      
      // generate each proxy method
      for (int i = 0; i < methods.length; i++)
      {
         ProxyMethod proxyMethod = new ProxyMethod(i);
         proxyMethod.assemble();
      }
      
      // complete the assembly
      classWriter.visitEnd();
      
      // get the byte array that represents the class file
      byte[] code = classWriter.toByteArray();
      
      return code;
   }
   
   /**
    * Assemble the <code>__initializeMethods(Method[])</code> method.  This is
    * a static method which is called once, by the {@link ProxyFactory},
    * immediately after loading the class.  It is assembled such that it will
    * throw an <code>IllegalStateException</code> if it is invoked again.
    */
   private void assembleInitializeMethods()
   {
      CodeVisitor cv =
         classWriter.visitMethod(ACC_PUBLIC | ACC_STATIC,
                                 INIT_METHODS_METHOD,
                                 "([Ljava/lang/reflect/Method;)V",
                                 null,
                                 null);
      
      // get the current value of the "ma" static field
      cv.visitFieldInsn(GETSTATIC,
                        proxyName,
                        "ma",
                        "[Ljava/lang/reflect/Method;");
      
      // if "ma" is still null, skip over the next set of instructions (which
      // otherwise throw IllegalStateException)
      Label jumpTo = new Label();
      cv.visitJumpInsn(IFNULL, jumpTo);
      
      // if we did not skip over this next set of instructions, the "ma" field
      // already was set;  it is invalid to call this method to set it again,
      // so throw an IllegalStateException
      cv.visitTypeInsn(NEW, "java/lang/IllegalStateException");
      cv.visitInsn(DUP);
      cv.visitLdcInsn("Proxy methods may not be set more than once");
      cv.visitMethodInsn(INVOKESPECIAL,
                         "java/lang/IllegalStateException",
                         "<init>",
                         "(Ljava/lang/String;)V");
      cv.visitInsn(ATHROW);
      
      // this is where we jump to if all is well...
      cv.visitLabel(jumpTo);
      
      // load the method's single argument
      cv.visitVarInsn(ALOAD, 0);
      
      // store the array of methods into the "ma" static field
      cv.visitFieldInsn(PUTSTATIC,
                        proxyName,
                        "ma",
                        "[Ljava/lang/reflect/Method;");
      
      // return
      cv.visitInsn(RETURN);
      
      // set max stack depth and local variables
      cv.visitMaxs(3, 1);
   }
   
   /**
    * Assemble the proxy class' single, default constructor.
    */
   private void assembleDefaultConstructor()
   {
      CodeVisitor cv =
         classWriter.visitMethod(ACC_PUBLIC,
                                 "<init>",
                                 "()V",
                                 null,
                                 null);
      
      // load implicit "this" variable
      cv.visitVarInsn(ALOAD, 0);
      
      // invoke superclass' default constructor
      cv.visitMethodInsn(INVOKESPECIAL, parentName, "<init>", "()V");
      
      // return from c'tor
      cv.visitInsn(RETURN);
      
      // set max stack depth and local variables
      cv.visitMaxs(1, 1);
   }
   
   /**
    * Assemble the proxy class' single constructor, which accepts and stores
    * the <code>InvocationHandler</code> to which method calls will be
    * dispatched.
    */
   private void assembleHandlerConstructor()
   {
      CodeVisitor cv =
         classWriter.visitMethod(ACC_PUBLIC,
                                 "<init>",
                                 "(Ljava/lang/reflect/InvocationHandler;)V",
                                 null,
                                 null);
      
      // load implicit "this" variable
      cv.visitVarInsn(ALOAD, 0);
      
      // invoke superclass' default constructor
      cv.visitMethodInsn(INVOKESPECIAL, parentName, "<init>", "()V");
      
      // load implicit "this" variable
      cv.visitVarInsn(ALOAD, 0);
      
      // load argument (InvocationHandler) passed to c'tor
      cv.visitVarInsn(ALOAD, 1);
      
      // store it in the "h" instance variable
      cv.visitFieldInsn(PUTFIELD,
                        proxyName,
                        "h",
                        "Ljava/lang/reflect/InvocationHandler;");
      
      // return from c'tor
      cv.visitInsn(RETURN);
      
      // set max stack depth and local variables
      cv.visitMaxs(2, 2);
   }
   
   /**
    * Helper class to assemble a proxy method.  The proxy method delegates its
    * invocation to an instance of <code>java.lang.reflect.Method</code> in
    * the <code>ma</code> array in the containing class.
    * <p>
    * Pseudo-code for the generated method is:
    * <pre>
    * public ReturnType methodName([ArgumentList])
    * [throws ExceptionList]
    * {
    *    try
    *    {
    *       return (ReturnType)
    *          [Handler].invoke(this,
    *                           ma[index],
    *                           [ArgumentList (as Object[])] or null);
    *    }
    *    [
    *    catch (Exception1 exc)
    *    {
    *       throw exc;
    *    }
    *    [
    *    catch (Exception2 exc)
    *    {
    *       throw exc;
    *    }
    *    ]
    *    ...
    *    [
    *    catch (ExceptionN exc)
    *    {
    *       throw exc;
    *    }
    *    ]]]
    *    catch (Throwable thr)
    *    {
    *       throw new UndeclaredThrowableException(thr);
    *    }
    * }
    * </pre>
    * <p>
    * where:
    * <ul>
    * <li><code>index</code> is an index into the enclosing class' array of
    *     delegate <code>Method</code> objects <code>ma</code>;
    * <li><code>Handler</code> is the "this" pointer if the proxy itself
    *     is the invocation handler, or the "h" instance field if a separate
    *     invocation handler is stored within the proxy.
    * </ul>
    * <p>
    * For methods which return <code>void</code>, the <code>return</code>
    * statement and cast to <code>ReturnType</code> are omitted.  In the event
    * of an empty argument list, an empty array of type <code>Object</code>
    * is passed to the <code>InvocationHandler.invoke</code> method.  If the
    * target method declares no exceptions, only the final catch block, which
    * catches <code>Throwable</code> is generated.  If the target method
    * explicitly declares that it <code>throws Throwable</code>, this catch
    * block is omitted.
    */
   private class ProxyMethod
   {
      /** Index of delegate method in containing class' array of Methods */
      private final Integer index;
      
      /** Object which describes method being proxied */
      private final Method method;
      
      /**
       * Constructor.
       * 
       * @param   index
       *          Index of delegate method in containing class' array of
       *          Methods.
       */
      ProxyMethod(int index)
      {
         this.index = new Integer(index);
         this.method = methods[index];
      }
      
      /**
       * Assemble the bytecode which comprises the proxy method.
       */
      void assemble()
      {
         int maxStackDepth = 0;
         
         String desc = Utils.makeMethodDescriptor(method);
         
         CodeVisitor cv =
            classWriter.visitMethod(ACC_PUBLIC,
                                    method.getName(),
                                    desc,
                                    Utils.getCaughtExceptionTypeNames(method),
                                    null);
         
         // labels which identify the start and end of the try block
         Label tryStart = new Label();
         Label tryEnd = new Label();
         
         // mark start of try block
         cv.visitLabel(tryStart);
         
         // load invocation handler from field "h" or as "this" if the proxy
         // itself is the invocation handler
         cv.visitVarInsn(ALOAD, 0);
         maxStackDepth++;
         if (!parentIsHandler)
         {
            cv.visitFieldInsn(GETFIELD,
                              proxyName,
                              "h",
                              "Ljava/lang/reflect/InvocationHandler;");
         }
         
         // load "this" object ref as the first parameter for later call to
         // InvocationHandler.invoke
         cv.visitVarInsn(ALOAD, 0);
         maxStackDepth++;
         
         // load target method from array of methods in static field "ma"
         cv.visitFieldInsn(GETSTATIC,
                           proxyName,
                           "ma",
                           "[Ljava/lang/reflect/Method;");
         maxStackDepth++;
         
         // target method is at index "index" in the "ma" array
         pushInt(cv, index);
         maxStackDepth++;
         cv.visitInsn(AALOAD);
         
         // create new array of Objects for method's arguments or push null
         // if there are no arguments
         Class<?>[] parmTypes = method.getParameterTypes();
         Integer numArgs = new Integer(parmTypes.length);
         if (numArgs > 0)
         {
            pushInt(cv, numArgs);
            cv.visitTypeInsn(ANEWARRAY, "java/lang/Object");
         }
         else
         {
            cv.visitInsn(ACONST_NULL);
         }
         
         // get ready to track parameters' effect on max stack depth
         int maxParmDepth = 0;
         
         // in preparation for marshaling our parameters, start the local
         // variable pool index at 1 (0 is reserved for the "this" reference)
         int localIndex = 1;
         
         // prepare arguments for InvocationHandler.invoke call
         for (int i = 0; i < parmTypes.length; i++)
         {
            TypeInfo parmInfo = Utils.getTypeInfo(parmTypes[i]);
            int width = parmInfo.getWidth();
            
            // duplicate array reference on top of stack
            cv.visitInsn(DUP);
            
            // push the target index in the array
            pushInt(cv, i);
            
            // load the value from the local variable pool onto the stack
            cv.visitVarInsn(parmInfo.getLoadInsn(), localIndex);
            
            // update parameter's maximum effect on stack depth
            maxParmDepth = Math.max(maxParmDepth, 2 + width);
            
            // if parameter is a primitive, it must be wrapped
            MethodInfo wrapInfo = parmInfo.getWrapInfo();
            if (wrapInfo != null)
            {
               cv.visitMethodInsn(INVOKESTATIC,
                                  wrapInfo.getClassName(),
                                  wrapInfo.getMethodName(),
                                  wrapInfo.getDescriptor());
            }
            
            // store Object parameter into array
            cv.visitInsn(AASTORE);
            
            // update the current index in the variable pool
            localIndex += width;
         }
         
         // track parameters' (if any) effect on max stack depth
         maxStackDepth += maxParmDepth;
         
         // call InvocationHandler.invoke
         cv.visitMethodInsn(INVOKEINTERFACE,
                            "java/lang/reflect/InvocationHandler",
                            "invoke",
                            "(Ljava/lang/Object;Ljava/lang/reflect/Method;[Ljava/lang/Object;)Ljava/lang/Object;");
         
         // determine method's return type
         Class<?> returnType = method.getReturnType();
         TypeInfo returnInfo = Utils.getTypeInfo(returnType);
         
         // cast return type to expected type, unless return type is void
         // (indicated by null cast name) or is java/lang/Object
         String castName = returnInfo.getCastName();
         if (castName != null && !"java/lang/Object".equals(castName))
         {
            cv.visitTypeInsn(CHECKCAST, returnInfo.getCastName());
         }
         
         // extract primitive from java wrapper type, if necessary
         MethodInfo unwrapInfo = returnInfo.getUnwrapInfo();
         if (unwrapInfo != null)
         {
            cv.visitMethodInsn(INVOKEVIRTUAL,
                               unwrapInfo.getClassName(),
                               unwrapInfo.getMethodName(),
                               unwrapInfo.getDescriptor());
         }
         
         // if return type is void, we have to pop the value returned by the
         // invocation handler off the stack before we return
         if (castName == null)
         {
            cv.visitInsn(POP);
         }
         
         // mark end of try block (includes next instruction)
         cv.visitLabel(tryEnd);
         
         // return;  even a void method requires a return instruction
         cv.visitInsn(returnInfo.getReturnInsn());
         
         // handle declared exceptions
         String[] exceptions = Utils.getCaughtExceptionTypeNames(method);
         boolean isThrowableDeclared = false;
         
         for (String exception : exceptions)
         {
            if ("java/lang/Throwable".equals(exception))
            {
               isThrowableDeclared = true;
               
               // if Throwable is declared, no need to catch it, just let it
               // be thrown
               continue;
            }
            
            Label excLabel = new Label();
            
            // mark start of exception catch block
            cv.visitLabel(excLabel);
            
            // store exception object ref into local variable, reload it, then
            // throw it
            cv.visitVarInsn(ASTORE, localIndex);
            cv.visitVarInsn(ALOAD, localIndex);
            cv.visitInsn(ATHROW);
            
            // register catch block
            cv.visitTryCatchBlock(tryStart,
                                  tryEnd,
                                  excLabel,
                                  exception);
         }
         
         if (!isThrowableDeclared)
         {
            Label catchThrowable = new Label();
            
            // mark start of unexpected throwable catch block
            cv.visitLabel(catchThrowable);
            
            // store exception object ref into local variable, wrap it within
            // UndeclaredThrowableException, and throw the latter
            cv.visitVarInsn(ASTORE, localIndex);
            cv.visitTypeInsn(NEW, "java/lang/reflect/UndeclaredThrowableException");
            cv.visitInsn(DUP);
            cv.visitVarInsn(ALOAD, localIndex);
            cv.visitMethodInsn(INVOKESPECIAL,
                               "java/lang/reflect/UndeclaredThrowableException",
                               "<init>",
                               "(Ljava/lang/Throwable;)V");
            cv.visitInsn(ATHROW);
            
            // register catch block
            cv.visitTryCatchBlock(tryStart,
                                  tryEnd,
                                  catchThrowable,
                                  "java/lang/Throwable");
         }
         
         // set max stack depth and local variables
         cv.visitMaxs(maxStackDepth, localIndex + 1);
      }
   }
   
   /**
    * Helper method to choose the most appropriate instruction with which to
    * push an integer value onto the stack.  The instruction used differs,
    * based upon the value of the integer.
    * 
    * @param   cv
    *          <code>CodeVisitor</code> used to write the instruction.
    * @param   value
    *          Integer value to be pushed.
    */
   private void pushInt(CodeVisitor cv, int value)
   {
      switch (value)
      {
         case 0:
            cv.visitInsn(ICONST_0);
            break;
         case 1:
            cv.visitInsn(ICONST_1);
            break;
         case 2:
            cv.visitInsn(ICONST_2);
            break;
         case 3:
            cv.visitInsn(ICONST_3);
            break;
         case 4:
            cv.visitInsn(ICONST_4);
            break;
         case 5:
            cv.visitInsn(ICONST_5);
            break;
         case -1:
            cv.visitInsn(ICONST_M1);
            break;
         default:
            if (value >= Byte.MIN_VALUE && value <= Byte.MAX_VALUE)
            {
               cv.visitIntInsn(BIPUSH, value);
            }
            else if (value >= Short.MIN_VALUE && value <= Short.MAX_VALUE)
            {
               cv.visitIntInsn(SIPUSH, value);
            }
            else
            {
               cv.visitLdcInsn(value);
            }
            break;
      }
   }
}
