/*
** Module   : ProxyFactory.java
** Abstract : Utility class to create and manage dynamic proxy classes.
**
** Copyright (c) 2010-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description------------------------
** 001 ECF 20101115 Created initial version.
** 002 ECF 20101214 Added variants of getProxy methods which do not accept
**                  InvocationHandler instances. These assume the proxy's
**                  superclass implement InvocationHandler.
** 003 ECF 20110101 Fixed CacheKey implementation to deal correctly with array
**                  hash codes. Fixed class-loading logic to use correct
**                  loader when cached class name is found.
** 004 ECF 20130120 Added isProxyClass method to test whether a class was
**                  generated by this factory; fix to enable nested interfaces
**                  to be used to define the proxy.
*/

package com.goldencode.proxy;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * A factory for dynamic proxy objects, which defines new classes using the
 * ASM bytecode engineering library (http://asm.ow2.org).  This class manages
 * a cache of proxy classes and implements the loading of these dynamically
 * generated classes using a custom class loader.  However, the work of
 * actually assembling the proxy classes is delegated to the {@link
 * ProxyAssembler}.  Please see the description of that class for the
 * rationale of this mechanism as an alternative to Java's standard, dynamic
 * proxy implementation.
 * <p>
 * The access points to this factory's functionality are the public, static
 * variants of the <code>getProxy</code> method, which allow the definition
 * of a new proxy class, based on a given superclass, list of interfaces to
 * implement, and an invocation handler to which method invocations are
 * dispatched.  Several additional options are available to tailor the proxy
 * class to specific needs.  Please see:
 * <ul>
 * <li>{@link #getProxy(Class, Class[], InvocationHandler)}
 * <li>{@link #getProxy(Class, Class[], boolean, Method[], InvocationHandler)}
 * <li>{@link #getProxy(Class, Class[])}
 * <li>{@link #getProxy(Class, Class[], boolean, Method[])}
 * </ul>
 * 
 * @author ECF
 */
public final class ProxyFactory
{
   /** Singleton instance of this factory class */
   private static final ProxyFactory singleton = new ProxyFactory();
   
   /** Cache of names for proxy classes which already have been assembled */
   private final Map<CacheKey, String> cache =
      new HashMap<CacheKey, String>();
   
   /** Custom class loader to load proxy classes */
   private final ProxyClassLoader loader = new ProxyClassLoader();
   
   /** Suffix to ensure unique proxy class names */
   private int nextSuffix = 0;
   
   /**
    * Private constructor to ensure singleton instance.
    */
   private ProxyFactory()
   {
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  Invocations of the methods in
    * those interfaces will be dispatched to the specified invocation handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * the same names, compatible signatures and return types, but different
    * lists of exception types, the first method discovered which has a return
    * type assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  Therefore, callers should take
    * care to specify methods with matching exception lists, if methods with 
    * the same name, signature, and compatible return types are specified
    * across interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   handler
    *          Handler to which to dispatch method invocations.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   @SuppressWarnings("unchecked")
   public static <T> T getProxy(Class<?> parent,
                                Class<?>[] interfaces,
                                InvocationHandler handler)
   {
      return (T) getProxy(parent, interfaces, true, null, handler);
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  Invocations of the methods in
    * those interfaces will be dispatched to the specified invocation handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * compatible signatures and return types, but different lists of exception
    * types, the first method discovered which has a return type which is
    * assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  The order in which methods are
    * checked is indeterminate.  Therefore, callers should be careful to
    * specify methods with matching exception lists, if methods with the same
    * name, signature, and compatible return types are specified across
    * interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   declaredOnly
    *          If <code>true</code>, only methods explicitly declared by the
    *          given interfaces are used;  if <code>false</code>, methods
    *          declared by the given interfaces and by any of their ancestor
    *          interfaces are used.
    * @param   doNotProxy
    *          An optional array of <code>Method</code> objects which are not
    *          to be implemented by the dynamically generated proxy.  This
    *          option should be used only to represent methods which already
    *          are implemented by the proxy class' superclass.
    * @param   handler
    *          Handler to which to dispatch method invocations.  This value
    *          must not be <code>null</code>.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if <code>handler</code> is <code>null</code>;
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   @SuppressWarnings("unchecked")
   public static <T> T getProxy(Class<?> parent,
                                Class<?>[] interfaces,
                                boolean declaredOnly,
                                Method[] doNotProxy,
                                InvocationHandler handler)
   {
      if (handler == null)
      {
         throw new IllegalArgumentException(
            "Invocation handler must not be null");
      }
      
      return (T) singleton.getProxyImpl(parent,
                                        interfaces,
                                        declaredOnly,
                                        doNotProxy,
                                        handler);
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  This method variant does not
    * accept an <code>InvocationHandler</code> instance.  It should be used
    * when the parent class is expected to fulfill the role of invocation
    * handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * compatible signatures and return types, but different lists of exception
    * types, the first method discovered which has a return type which is
    * assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  The order in which methods are
    * checked is indeterminate.  Therefore, callers should be careful to
    * specify methods with matching exception lists, if methods with the same
    * name, signature, and compatible return types are specified across
    * interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class, which must implement the
    *          <code>InvocationHandler</code> interface.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   @SuppressWarnings("unchecked")
   public static <T> T getProxy(Class<? extends InvocationHandler> parent,
                                Class<?>[] interfaces)
   {
      return (T) getProxy(parent, interfaces, true, null);
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  This method variant does not
    * accept an <code>InvocationHandler</code> instance.  It should be used
    * when the parent class is expected to fulfill the role of invocation
    * handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * compatible signatures and return types, but different lists of exception
    * types, the first method discovered which has a return type which is
    * assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  The order in which methods are
    * checked is indeterminate.  Therefore, callers should be careful to
    * specify methods with matching exception lists, if methods with the same
    * name, signature, and compatible return types are specified across
    * interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class, which must implement the
    *          <code>InvocationHandler</code> interface.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   declaredOnly
    *          If <code>true</code>, only methods explicitly declared by the
    *          given interfaces are used;  if <code>false</code>, methods
    *          declared by the given interfaces and by any of their ancestor
    *          interfaces are used.
    * @param   doNotProxy
    *          An optional array of <code>Method</code> objects which are not
    *          to be implemented by the dynamically generated proxy.  This
    *          option should be used only to represent methods which already
    *          are implemented by the proxy class' superclass.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   @SuppressWarnings("unchecked")
   public static <T> T getProxy(Class<? extends InvocationHandler> parent,
                                Class<?>[] interfaces,
                                boolean declaredOnly,
                                Method[] doNotProxy)
   {
      return (T) singleton.getProxyImpl(parent,
                                        interfaces,
                                        declaredOnly,
                                        doNotProxy,
                                        null);
   }
   
   /**
    * Indicate whether the given class represents a proxy class generated by
    * this factory.
    * 
    * @param   clazz
    *          The class to be tested.
    * 
    * @return  <code>true</code> if the given class represents a proxy class
    *          generated by this factory, else <code>false</code>.
    */
   public static boolean isProxyClass(Class<?> clazz)
   {
      return singleton.loader.equals(clazz.getClassLoader());
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  Invocations of the methods in
    * those interfaces will be dispatched to the specified invocation handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * compatible signatures and return types, but different lists of exception
    * types, the first method discovered which has a return type which is
    * assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  The order in which methods are
    * checked is indeterminate.  Therefore, callers should be careful to
    * specify methods with matching exception lists, if methods with the same
    * name, signature, and compatible return types are specified across
    * interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class.  If <code>handler</code> is
    *          <code>null</code>, this class is expected to implement the
    *          <code>InvocationHandler</code> interface.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   declaredOnly
    *          If <code>true</code>, only methods explicitly declared by the
    *          given interfaces are used;  if <code>false</code>, methods
    *          declared by the given interfaces and by any of their ancestor
    *          interfaces are used.
    * @param   doNotProxy
    *          An optional array of <code>Method</code> objects which are not
    *          to be implemented by the dynamically generated proxy.  This
    *          option should be used only to represent methods which already
    *          are implemented by the proxy class' superclass.
    * @param   handler
    *          Handler to which to dispatch method invocations.  If
    *          <code>null</code>, the dynamic proxy instance itself will be
    *          used as the invocation handler.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   @SuppressWarnings("unchecked")
   private <T> T getProxyImpl(Class<?> parent,
                              Class<?>[] interfaces,
                              boolean declaredOnly,
                              Method[] doNotProxy,
                              InvocationHandler handler)
   {
      Class<?> clazz = null;
      T proxy = null;
      boolean proxyIsHandler = (handler == null);
      
      if (doNotProxy == null)
      {
         doNotProxy = new Method[0];
      }
      CacheKey key = new CacheKey(parent,
                                  interfaces,
                                  declaredOnly,
                                  doNotProxy);
      
      synchronized (this)
      {
         String proxyName = cache.get(key);
         
         try
         {
            if (proxyName != null)
            {
               clazz = Class.forName(proxyName, false, loader);
            }
            else
            {
               String parentName = parent.getName();
               proxyName = makeProxyClassName(parentName);
               Method[] methods = gatherMethods(interfaces,
                                                declaredOnly,
                                                doNotProxy);
               
               // assemble code for the proxy class
               ProxyAssembler asm = new ProxyAssembler(proxyName,
                                                       parentName,
                                                       interfaces,
                                                       methods,
                                                       proxyIsHandler);
               byte[] code = asm.assembleClass();
               
               /*
               // debug code
               FileOutputStream fos;
               try
               {
                  int pos = proxyName.lastIndexOf(".");
                  String fname = (pos >= 0
                                  ? proxyName.substring(pos + 1)
                                  : proxyName);
                  System.out.println("Proxy class file written to: " + fname);
                  fos = new FileOutputStream(fname);
                  fos.write(code);
                  fos.close();
               }
               catch (IOException exc)
               {
                  exc.printStackTrace();
               }
               */
               
               // load the proxy class with a custom class loader
               clazz = loader.loadClass(proxyName, code);
               
               // call the proxy class' __initializeMethods method, to assign
               // the array of Method objects instances will pass to their
               // invocation handlers
               Method initMethsMethod =
                  clazz.getDeclaredMethod(ProxyAssembler.INIT_METHODS_METHOD,
                                          Method[].class);
               initMethsMethod.invoke(clazz, (Object) methods);
               
               // remember that we've assembled a proxy for the current set of
               // criteria
               cache.put(key, proxyName);
            }
            
            if (proxyIsHandler)
            {
               // use default constructor
               proxy = (T) clazz.newInstance();
            }
            else
            {
               // use constructor which stores separate handler
               Constructor<?> ctor =
                  clazz.getDeclaredConstructor(InvocationHandler.class);
               proxy = (T) ctor.newInstance(handler);
            }
         }
         catch (ClassNotFoundException exc)
         {
            throwInternalError(exc);
         }
         catch (NoSuchMethodException exc)
         {
            throwInternalError(exc);
         }
         catch (InvocationTargetException exc)
         {
            throwInternalError(exc);
         }
         catch (IllegalAccessException exc)
         {
            throwInternalError(exc);
         }
         catch (InstantiationException exc)
         {
            throwInternalError(exc);
         }
      }
      
      return proxy;
   }
   
   /**
    * Throw an internal error, with the given exception as its root cause.
    * 
    * @param   exc
    *          Root cause of error.
    */
   private void throwInternalError(Exception exc)
   {
      InternalError error = new InternalError(exc.getMessage());
      error.initCause(exc);
      
      throw error;
   }
   
   /**
    * Given the name of the proxy's superclass, generate a unique name of the
    * proxy class, which can be used to load the class in the custom, proxy
    * class loader.  This name differs from the internal name of the proxy
    * class, in that it uses the dot (</code>.</code>) character as its
    * separator, whereas the internal name uses the forward slash
    * (<code>/</code>).
    * 
    * @param   parentName
    *          Fully qualified name of parent class.
    * 
    * @return  Unique name for the proxy class, qualified by the
    *          <code>com.goldencode.proxy</code> package name.
    */
   private String makeProxyClassName(String parentName)
   {
      StringBuilder buf = new StringBuilder("com.goldencode.proxy.");
      String name = parentName.replaceAll("\\.", "_");
      buf.append(name);
      buf.append("__Proxy");
      buf.append(nextSuffix++);
      
      return buf.toString();
   }
   
   /**
    * Gather the final list of methods to be implemented explicitly by the
    * dynamic proxy class.  The list is determined by iterating through the
    * methods declared by the given interfaces (and optionally, by their
    * ancestor interfaces), omitting those explicitly identified as not to be
    * proxied.  In the event the same method name and parameter signatures are
    * found across interfaces, any potential conflicts are resolved by using
    * only that method (if any) whose return type is assignable from all
    * other, potentially conflicting methods.  In the event any such conflict
    * cannot be resolved, an exception is thrown.
    * 
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   declaredOnly
    *          If <code>true</code>, only methods explicitly declared by the
    *          given interfaces are used;  if <code>false</code>, methods
    *          declared by the given interfaces and by any of their ancestor
    *          interfaces are used.
    * @param   doNotProxy
    *          An optional array of <code>Method</code> objects which are not
    *          to be implemented by the dynamically generated proxy.  This
    *          option should be used only to represent methods which already
    *          are implemented by the proxy class' superclass.
    * 
    * @return  The final array of methods to be implemented explicitly by the
    *          dynamic proxy class.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   private Method[] gatherMethods(Class<?>[] interfaces,
                                  boolean declaredOnly,
                                  Method[] doNotProxy)
   {
      // create a set of methods which are to be omitted from proxy class
      Set<Method> omit = new HashSet<Method>();
      for (Method method : doNotProxy)
      {
         omit.add(method);
      }
      
      // map signatures (w/o return type) to sets of unique methods
      Map<String, Set<Method>> sig2Methods =
         new LinkedHashMap<String, Set<Method>>();
      
      // list of methods which survive screening process
      List<Method> survivors = new ArrayList<Method>();
      
      // collect list of methods declared by array of interfaces
      for (Class<?> iface : interfaces)
      {
         Method[] toBeProxied = (declaredOnly
                                 ? iface.getDeclaredMethods()
                                 : iface.getMethods());
         for (Method method : toBeProxied)
         {
            if (omit.contains(method))
            {
               continue;
            }
            
            StringBuilder buf = new StringBuilder(method.getName());
            buf.append(Utils.makeMethodSignatureDescriptor(method));
            String sig = buf.toString();
            Set<Method> methods = sig2Methods.get(sig);
            if (methods == null)
            {
               methods = new LinkedHashSet<Method>();
               sig2Methods.put(sig, methods);
            }
            methods.add(method);
         }
      }
      
      // walk through all groups of methods;  collect survivors
      for (Map.Entry<String, Set<Method>> entry : sig2Methods.entrySet())
      {
         Set<Method> methods = entry.getValue();
         Method survivor = identifyBestMethod(methods);
         survivors.add(survivor);
      }
      
      Method[] methods = survivors.toArray(new Method[survivors.size()]);
      
      return methods;
   }
   
   /**
    * Given a set of methods, identify the best method to implement in the
    * dynamic proxy.  If the set contains only a single method, it is used.
    * In the event the set contains multiple methods, the method whose return
    * type is assignable from all others in the set is chosen.  If no such
    * method exists, an exception is thrown.
    * 
    * @param   methods
    *          A set of methods which share a name and parameter signature,
    *          but not necessarily the same return type.
    * 
    * @return  The best method to implement in the dynamic proxy, according to
    *          the above criteria.
    * 
    * @throws  IllegalArgumentException
    *          if multiple methods exist in the set, but none has a return
    *          type that is assignable from all the others.
    */
   private Method identifyBestMethod(Set<Method> methods)
   {
      if (methods.size() == 1)
      {
         return methods.iterator().next();
      }
      
      outerLoop:
      for (Iterator<Method> outer = methods.iterator(); outer.hasNext(); )
      {
         Method mOuter = outer.next();
         Class<?> rtOuter = mOuter.getReturnType();
         
         for (Iterator<Method> inner = methods.iterator(); inner.hasNext(); )
         {
            Method mInner = inner.next();
            Class<?> rtInner = mInner.getReturnType();
            if (!rtOuter.isAssignableFrom(rtInner))
            {
               continue outerLoop;
            }
         }
         
         return mOuter;
      }
      
      throw new IllegalArgumentException(
         "Methods have same signature, but incompatible return types: " +
         methods);
   }
   
   /**
    * A cache key which takes into account a proxy class' parent, implemented
    * interfaces, whether only declared methods are implemented, and list of
    * methods to be omitted from the proxy.
    * <p>
    * If all the above criteria match between two keys, the keys are
    * equivalent;  else they are not.
    * <p>
    * Instances of this class are used in a cache, to ensure redundant proxy
    * classes are not assembled and loaded.
    */
   private static class CacheKey
   {
      /** Parent class of proxy class */
      private final Class<?> parent;
      
      /** Interfaces implemented by proxy class */
      private final Class<?>[] interfaces;
      
      /** Flag indicating only declared methods of interfaces are proxied */
      private final boolean declaredOnly;
      
      /** Methods which are not to be proxied */
      private final Method[] doNotProxy;
      
      /**
       * Constructor.
       * 
       * @param   parent
       *          Parent class of proxy class.
       * @param   interfaces
       *          Interfaces implemented by proxy class.
       * @param   declaredOnly
       *          Flag indicating only declared methods of interfaces are
       *          proxied.
       * @param   doNotProxy
       *          Methods which are not to be proxied.  May be an empty array,
       *          but must not be <code>null</code>.
       */
      CacheKey(Class<?> parent,
               Class<?>[] interfaces,
               boolean declaredOnly,
               Method[] doNotProxy)
      {
         this.parent = parent;
         this.interfaces = interfaces;
         this.declaredOnly = declaredOnly;
         this.doNotProxy = doNotProxy;
      }
      
      /**
       * Return a well-distributed hash code which is consistent with this
       * class' implementation of {@link #equals(Object)}.
       * 
       * @return  Hash code.
       */
      public int hashCode()
      {
         int result = 17;
         
         result = 37 * result + parent.hashCode();
         result = 37 * result + Arrays.hashCode(interfaces);
         result = 37 * result + (declaredOnly
                                 ? Boolean.TRUE.hashCode()
                                 : Boolean.FALSE.hashCode());
         result = 37 * result + Arrays.hashCode(doNotProxy);
         
         return result;
      }
      
      /**
       * Determine whether this object is equivalent to the given object.
       * <p>
       * Two cache keys are considered equivalent if their various fields are
       * equivalent.
       * 
       * @param   o
       *          Another object to test for equivalence with this one.
       * 
       * @return  <code>true</code> if <code>o</code> is another instance of
       *          this class with equivalent fields, else <code>false</code>.
       */
      public boolean equals(Object o)
      {
         if (!(o instanceof CacheKey))
         {
            return false;
         }
         
         CacheKey that = (CacheKey) o;
         
         return (this.parent.equals(that.parent) &&
                 Arrays.equals(this.interfaces, that.interfaces) &&
                 this.declaredOnly == that.declaredOnly &&
                 Arrays.equals(this.doNotProxy, that.doNotProxy));
      }
   }
   
   /**
    * A custom class loader which loads dynamically generated proxy classes.
    * The classes are assembled by the {@link ProxyAssembler} in memory.  Each
    * is represented as an array of bytes which conforms to the Java class
    * file format as specified by <u>The Java Virtual Machine Specification,
    * Second Edition</u> by Tim Lindholm and Frank Yellin.
    * <p>
    * A class is loaded by invoking the <code>loadClass(String, byte[])</code>
    * method, passing it the fully qualified name of the proxy class, and the
    * byte code array representing its class file.
    */
   private final static class ProxyClassLoader
   extends ClassLoader
   {
      /** Thread-local, temporary association of a class' name to its data */
      private final ThreadLocal<Data> local = new ThreadLocal<Data>();
      
      /**
       * Default constructor.
       */
      ProxyClassLoader()
      {
         super();
      }
      
      /**
       * Overrides the parent's implementation to find the class file data
       * to load the dynamically generated proxy class with the given name.
       * 
       * @param   name
       *          Name of the proxy class to be loaded.
       * 
       * @return  Dynamically loaded proxy class.
       * 
       * @throws  ClassNotFoundException
       *          if no thread-local class file data is found, or if its name
       *          does not match the given class name.
       */
      @Override
      public Class<?> findClass(String name)
      throws ClassNotFoundException
      {
         Data data = local.get();
         if (data == null || !name.equals(data.name))
         {
            throw new ClassNotFoundException(name);
         }
         
         byte[] code = data.code;
         Class<?> clazz = defineClass(name, code, 0, code.length);
         
         return clazz;
      }
      
      /**
       * Store a piece of thread-local data which associates the given proxy
       * class name with the byte array representing its class file.  This
       * information is stored temporarily, while the work of loading the
       * class is delegated to the <code>loadClass(String, boolean)</code>
       * method, then the thread-local data is immediately removed.
       * 
       * @param   name
       *          Name of the proxy class to be loaded.
       * @param   code
       *          Byte array representing the proxy's class file.
       * 
       * @return  The proxy class.
       * 
       * @throws  ClassNotFoundException
       *          if something unexpected goes horribly wrong with the
       *          thread-local data mechanism.
       */
      Class<?> loadClass(String name, byte[] code)
      throws ClassNotFoundException
      {
         try
         {
            Data data = new Data();
            data.name = name;
            data.code = code;
            local.set(data);
            
            return loadClass(name, true);
         }
         finally
         {
            local.remove();
         }
      }
      
      /**
       * Association between a proxy's class name and its class file data.
       * Used transiently during class loading.
       */
      private static class Data
      {
         /** Class name */
         private String name;
         
         /** Byte array representing class file */
         private byte[] code;
      }
   }
}
